# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from abc import abstractmethod, ABCMeta

from ubuntu_system_tests.helpers import autopilot
from ubuntu_system_tests.helpers import wait_until
from ubuntu_system_tests.helpers import processes
from ubuntu_system_tests.helpers.click import get_click_app_identifier
from ubuntu_system_tests.helpers.processes import ensure_application_closed
from ubuntu_system_tests.helpers.scopes.apps import (
    launch_application_from_apps_scope)
from ubuntu_system_tests.helpers.unity8 import (
    launch_application_from_launcher)


class Application:
    __metaclass__ = ABCMeta

    def __init__(self, app_name, app_click, app_id=None):
        self.app_name = app_name
        self.app_click = app_click
        self.app_id = app_id

    def launch_from_scope(self, return_proxy=True):
        """
        Launch app from apps scope
        """
        launch_application_from_apps_scope(self.app_name)
        if return_proxy:
            return self.proxy_object

    def launch_from_launcher(self, return_proxy=True):
        """Drag out the launcher and tap the application icon to start it."""
        launch_application_from_launcher(self.app_id)
        if return_proxy:
            return self.proxy_object

    @property
    def proxy_object(self):
        wait_until(self.is_running)
        return self.get_proxy_object()

    def close(self):
        pass

    def stop(self):
        if self.app_id is None:
            self.app_id = get_click_app_identifier()
        if self.is_running():
            ensure_application_closed(self.app_id)

    @abstractmethod
    def is_running(self):
        pass

    @abstractmethod
    def get_proxy_object(self):
        pass


class WebApplication(Application):

    def get_proxy_object(self):
        """To be implemented"""
        pass

    @abstractmethod
    def get_main_window(self, proxy_object):
        pass


class QmlApplication(Application):

    def __init__(self, app_name, app_click, app_process_name, app_id=None):
        self.app_process_name = app_process_name
        super().__init__(app_name, app_click, app_id)

    def get_proxy_object(self):
        return autopilot.get_proxy_object_for_existing_app(
            self.app_process_name)

    def is_running(self):
        """ Indicate if the app is currently running """
        return processes.is_process_running(self.app_process_name)


class QmlSceneApplication(Application):

    def __init__(self, app_name, app_click, qml_file, app_id=None):
        self.qml_file = qml_file
        super().__init__(app_name, app_click, app_id)

    def get_proxy_object(self):
        """Return the proxy object of the app."""
        return autopilot.get_proxy_object_for_existing_qmlscene_process(
            self.qml_file)

    def is_running(self):
        """ Indicate if the app is currently running """
        return processes.is_qmlscene_running_with_qmlfile(self.qml_file)
