# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers.observer import Observer
from ubuntu_system_tests.helpers.threads import StoppingThread

logger = logging.getLogger(__name__)


class SimPinObserver(Observer):
    """Observer object to dismiss sim pin entry screens."""

    def __init__(self, stop, result, sims):
        """
        :param stop: Event object to indicate when to stop observing.
        :param result: Event object to indicate when required sim pin screens
                       have been dismissed.
        :param sims: List of locked sim indexes that require sim pin
                     notification.
        """
        super().__init__(stop, result)
        self.sims = sims

    def start(self):
        """Start the observer."""
        super().start()
        logger.info('{} sim(s) locked.'.format(len(self.sims)))
        self.dismiss_sim_pin_screens()

    def _cleanup(self):
        """Cleanup actions before stopping."""
        if not self.result.is_set():
            logger.info(
                '{c} out of {e} expected sim pin entry dialogs '
                'diaplayed.'.format(c=self.count, e=len(self.sims)))

    def dismiss_sim_pin_screens(self):
        """Dismiss each required sim pin screen."""
        self.count = 0
        for sim in self.sims:
            # Wait for sim pin notification to be displayed.
            n_list = context.shared.unity.main_window.get_notifications_list()
            notification = n_list.get_sim_unlock_notification(
                sim, timeout=30, stop_event=self._stop)
            if notification:
                # notification is found, so dimiss it and move onto next
                notification.get_lockscreen().dismiss()
                self.count += 1
            elif self._stop.is_set():
                # Stop request is set, so just exit immediately.
                return
        self.result.set()
        self.stop()


def _start_sim_pin_observer(stop, result, sims):
    """
    Start a sim pin notification observer.

    :param stop: Event object to signal when to stop observing.
    :param result: Event object to signal when required number of notifications
                   are observed.
    :param sims: List of locked sims requiring pin input.

    """
    observer = SimPinObserver(stop, result, sims)
    observer.start()


def start_sim_pin_observer(result, sims):
    """
    Start an observer thread to dismiss sim pin notifications.

    :param result: Event object to signal when required number of notifications
                   are observed.
    :param sims: List of locked sims requiring pin input.
    :return: StoppingThread object which should be stopped by caller when
             required.

    """
    launch_thread = StoppingThread(
        target=_start_sim_pin_observer, args=[result, sims])
    launch_thread.start()
    return launch_thread
