# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import signal
import subprocess

import psutil

from unity8.process_helpers import get_job_status

from ubuntu_system_tests.helpers import wait_until

logger = logging.getLogger(__name__)


def is_process_running(pname):
    try:
        return bool(get_process_ids(pname))
    except ValueError:
        return False


def _query_pids_for_process(process_name):
    pids = [process.pid for process in psutil.process_iter()
            if process.name() == process_name]

    if not pids:
        raise ValueError('Process \'{}\' not running'.format(process_name))
    return pids


def get_process_id(pname):
    """
    Return PID for the requested process.

    :param pname: name of the process to get PID of.
    """
    pids = _query_pids_for_process(process_name=pname)
    if len(pids) > 1:
        raise ValueError(
            'More than one PID exists for process \'{}\''.format(pname))

    return pids[0]


def get_process_ids(pname):
    """Return PIDs of the requested process, in case there are multiple."""
    return _query_pids_for_process(process_name=pname)


def stop_process(pname):
    if is_process_running(pname):
        [process.terminate() for process in psutil.process_iter() if
         process.name() == pname]


def start_application(app_name, *args):
    subprocess.call(['ubuntu-app-launch', app_name] + list(args))


def ensure_application_closed(app_id):
    subprocess.call(['ubuntu-app-stop', app_id])


def wait_for_job_spawned(name, timeout=10):
    """Wait until the named job is in spawned state."""
    spawned = wait_until(
        lambda: 'start/spawned' in get_job_status(name), timeout=timeout)
    if not spawned:
        raise RuntimeError(
            'Job {} not in start/spwaned state within timeout {}'.format(
                name, timeout))


def is_qmlscene_running_with_qmlfile(qmlfile):
    """ Indicate if the qmlscene is running the qml file qmlfile """
    pline = subprocess.check_output('ps -ef | grep qmlscene', shell=True,
                                    universal_newlines=True)
    return qmlfile in pline


def get_qmlscene_process_id(qmlfile):
    """ Retrieve the process id for the qmlscene running the file qmlfile """
    pline = subprocess.check_output(
        'ps -ef | grep qmlscene | grep {}'.format(qmlfile), shell=True,
        universal_newlines=True)
    return int(pline.split()[1])


def stop_qmlscene_process(qmlfile):
    """ Stop the qmlscene process running the file qmlfile """
    if is_qmlscene_running_with_qmlfile(qmlfile):
        pid = get_qmlscene_process_id(qmlfile)
        os.kill(pid, signal.SIGKILL)


def get_process_args(pid):
    """ Retrieve the command line used to start the process """
    return subprocess.check_output('xargs -0 < /proc/{}/cmdline'.format(pid),
                                   shell=True, universal_newlines=True)
