# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess

from unity8.sensors import FakePlatformSensors as Unity8Sensors

from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers.power import write_to_powerd_fifo
from ubuntu_system_tests.helpers.testbed import run_command_with_sudo as sudo

DELAY_MS = 10
ACCEL_SENSOR_RESOLUTION = 0.1
ACCEL_SENSOR_MAX = 1000
ACCEL_SENSOR_MIN = 0
LIGHT_SENSOR_RESOLUTION = 1
LIGHT_SENSOR_MAX = 1000
LIGHT_SENSOR_MIN = 0
LIGHT_SENSOR_MID = (LIGHT_SENSOR_MAX + LIGHT_SENSOR_MIN) // 2
ORIENTATION_TOP_UP = 'top_up'
ORIENTATION_TOP_DOWN = 'top_down'
ORIENTATION_LEFT_UP = 'left_up'
ORIENTATION_RIGHT_UP = 'right_up'
SENSOR_OVERRIDE = 'UBUNTU_PLATFORM_API_TEST_OVERRIDE=sensors'


def get_fake_sensors_setup_commands():
    """Return list of commands required to configure the fake sensors."""
    cmds = ['create proximity',
            'create accel {} {} {}'.format(
                ACCEL_SENSOR_MIN, ACCEL_SENSOR_MAX, ACCEL_SENSOR_RESOLUTION),
            'create light {} {} {}'.format(
                LIGHT_SENSOR_MIN, LIGHT_SENSOR_MAX, LIGHT_SENSOR_RESOLUTION),
            '{} proximity far'.format(DELAY_MS),
            '{} light {}'.format(DELAY_MS, LIGHT_SENSOR_MID)]
    return cmds


def are_fake_sensors_enabled_for_process(pid):
    """Return True if fake sensors enabled for specified process."""
    if not os.path.exists('/tmp/sensor-fifo-{}'.format(pid)):
        return False

    try:
        output = sudo('cat /proc/{pid}/environ'.format(pid=pid)).decode()
    except subprocess.CalledProcessError:
        return False
    return output.find(SENSOR_OVERRIDE) >= 0


def _set_light_level(level, wait=1):
    """Set light sensor to specified value.
    :param value: Light level value.
    :param wait: How long to wait in ms before applying value.
    """
    cmd = '{} light {}'.format(wait, level)
    write_to_powerd_fifo([cmd])


def _set_proximity(value, wait=1):
    """Set proximity sensor to specified value.
    :param value: Either 'near' or 'far'.
    :param wait: How long to wait in ms before applying value.
    """
    cmd = '{} proximity {}'.format(wait, value)
    write_to_powerd_fifo([cmd])


def set_proximity_near(wait=1):
    """Set proximity sensor to near.
    :param wait: How long to wait in ms before applying value.
    """
    _set_proximity('near', wait)


def set_proximity_far(wait=1):
    """Set proximity sensor to far.
    :param wait: How long to wait in ms before applying value.
    """
    _set_proximity('far', wait)


def set_light_level_bright(wait=1):
    """Set light sensor to bright.
    :param wait: How long to wait in ms before applying value.
    """
    _set_light_level(LIGHT_SENSOR_MAX, wait)


def set_light_level_dark(wait=1):
    """Set light sensor to dark.
    :param wait: How long to wait in ms before applying value.
    """
    _set_light_level(LIGHT_SENSOR_MIN, wait)


def set_light_level_mid(wait=1):
    """Set light sensor to mid-range value.
    :param wait: How long to wait in ms before applying value.
    """
    _set_light_level(LIGHT_SENSOR_MID, wait)


def set_orientation(orientation):
    """Set accelerometer sensor orientation to specified value.
    :param orientation: Orientation to set.
    """
    Unity8Sensors(context.shared.get('unity_pid')).set_orientation(orientation)


def set_orientation_top_up():
    """Set accelerometer sensor orientation to top edge up."""
    set_orientation(ORIENTATION_TOP_UP)


def set_orientation_top_down():
    """Set accelerometer sensor orientation to top edge down."""
    set_orientation(ORIENTATION_TOP_DOWN)


def set_orientation_left_up():
    """Set accelerometer sensor orientation to left edge up."""
    set_orientation(ORIENTATION_LEFT_UP)


def set_orientation_right_up():
    """Set accelerometer sensor orientation to right edge up."""
    set_orientation(ORIENTATION_RIGHT_UP)
