#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import subprocess
import sys

from argparse import ArgumentParser
from collections import namedtuple

from ubuntu_system_tests.host import (
    commands as cmds,
    debian,
    results,
)
from ubuntu_system_tests.common import (
    clean_dir,
    config,
)

logger = logging.getLogger(__name__)

Option = namedtuple('Option', ['args', 'kwargs'])
OPTION_CACHE_DEPS = Option(
    ['--cache-deps', '-d'],
    {'action': 'store_true',
     'help': 'Cache dependencies on the device so they don\'t have to be '
     'downloaded for every test run.',
     'default': False})
OPTION_CONFIG = Option(
    ['--config', '-c'],
    {'help': 'The config file to use for this run.',
     'default': 'ubuntu-system-tests.conf'})
OPTION_CONFIG_SECTION = Option(
    ['--config-section', '-e'],
    {'help': 'The section name to use in config file.',
     'default': config.KEY_DEFAULT})
OPTION_DIST_UPGRADE = Option(
    ['--dist-upgrade', '-p'],
    {'action': 'store_true',
     'help': 'Perform a dist-upgrade on the device to install latest '
     'available updates from archive.',
     'default': False})
OPTION_INSTALL_TESTS = Option(
    ['--install-tests', '-i'],
    {'action': 'store_true',
     'help': 'Install all dependencies permanently in the device instead '
     'of in the /tmp directory. This includes all production and test '
     'dependencies.',
     'default': False})
OPTION_NETWORK_READY = Option(
    ['--network-ready', '-n'],
    {'action': 'store_true',
     'help': 'Assume that the network connection is already setup.',
     'default': False})
OPTION_SILENT = Option(
    ['--silent', '-s'],
    {'action': 'store_true',
     'help': 'Do not prompt for missing user config options, use default '
     'values instead.',
     'default': False})
OPTION_SUITES = Option(
    ['suites'],
    {'nargs': '*',
     'help': 'The list of test suites to run. If left blank then all tests '
             'from ubuntu_system_tests.tests will be run.'})
OPTION_TOUCH_VISUALIZATION = Option(
    ['--touch-visualization', '-t'],
    {'action': 'store_true',
     'help': 'Enable visualization of touch points on screen.',
     'default': False})
OPTION_UPDATE_APT = Option(
    ['--update-apt', '-u'],
    {'action': 'store_true',
     'help': 'Update apt lists before installing the dependencies.',
     'default': False})
OPTION_VERBOSE = Option(
    ['--verbose', '-v'],
    {'action': 'store_true',
     'help': 'Provide verbose logging.',
     'default': False})

COMMON_ARGS = [
    OPTION_CONFIG,
    OPTION_CONFIG_SECTION,
    OPTION_NETWORK_READY,
    OPTION_SILENT,
    OPTION_VERBOSE]
RUN_ARGS = [
    OPTION_SUITES,
    OPTION_TOUCH_VISUALIZATION]
SETUP_ARGS = [
    OPTION_CACHE_DEPS,
    OPTION_DIST_UPGRADE,
    OPTION_INSTALL_TESTS,
    OPTION_UPDATE_APT
]

COMMAND_RUN = 'run'
COMMAND_SETUP = 'setup'


def main(argv=None):
    args = _parse_arguments(argv or sys.argv[1:])
    if args.command not in [COMMAND_RUN, COMMAND_SETUP]:
        logger.warning(
            'Command option must be either "{r}" or "{s}". Use -h option for '
            'further help.'.format(r=COMMAND_RUN, s=COMMAND_SETUP))
        return

    if args.command == COMMAND_RUN:
        return run_system_tests(args)
    else:
        return setup_system_tests(args)


def _add_options(parser, options):
    """Add the specified list of options to the option parser object."""
    for option in options:
        parser.add_argument(*option.args, **option.kwargs)


def _parse_arguments(argv=None):
    """Parse the provided arguments and return in Namespace object."""
    parser = ArgumentParser(
        'ubuntu-system-tests',
        description='Autopilot system test suite for Ubuntu Touch devices. '
        'Allows devices to be setup for testing and then used to execute '
        'suites of autopilot tests from ubuntu-system-tests or other '
        'projects. For further help use -h option with required command.')
    subparsers = parser.add_subparsers(
        dest='command',
        help='Command options')
    run = subparsers.add_parser(
        COMMAND_RUN,
        help='Run tests on the device',
        description='Executes a list of tests on a connected device once it '
        'has been setup using the setup command. If no tests are specified '
        'to be run then all tests from ubuntu_system_tests.tests will be '
        'executed.')
    _add_options(run, COMMON_ARGS + RUN_ARGS)
    setup = subparsers.add_parser(
        COMMAND_SETUP,
        help='Setup device ready for testing',
        description='Sets up the device ready for testing. This can include '
        'performing apt-get update and dist-upgrade operations. Also it is '
        'possible to install apt-cacher-ng to allow downloaded dependencies '
        'to be cached on the device so they don\'t need to be downloaded for '
        'every test run. If no arguments are specified then the default '
        'operation will be to perform apt-get update, dist-upgrade and '
        'install apt-cacher-ng. If any single option is specified then only '
        'this action will be performed.')
    _add_options(setup, COMMON_ARGS + SETUP_ARGS)
    return parser.parse_args(args=argv)


def setup_system_tests(args):
    """Validate arguments and execute required setup commands on device."""
    if ((args.cache_deps or args.dist_upgrade or args.install_tests) and
            not args.update_apt):
        # An install operation is requested without update_apt, so set
        # update_apt to True.
        args.update_apt = True
    elif not (args.cache_deps or args.dist_upgrade or args.update_apt):
        # There were no optional args specified.
        # In this case apply the default actions, which are:
        # update_apt, cache_deps and dist_upgrade
        args.cache_deps = args.dist_upgrade = args.update_apt = True
    return run_autopkg_setup(args)


def run_system_tests(args):
    """Run system tests in a device connected via USB and report the results.
    :param args: command line options.
    :returns: 0 when successful, other value when failed.

    """
    print_warnings()
    config_stack = _prepare_config(args)

    enable_autopilot_introspection(config_stack)
    return_code = run_autopkg_test(config_stack, args)

    if return_code <= 6:
        # Return codes higher than 6 mean that there was a problem with adt or
        # the test bed, not a failure in the tests. So there will be no file
        # with results.
        generate_user_report(config_stack)

    return return_code


def _prepare_config(args):
    """Get config object for specified arguments and clear previous results."""
    config_stack = _get_config_stack(
        args.config, args.config_section, vars(args).get('silent', False))
    clean_output_directory(config_stack)
    return config_stack


def get_camera_warning():
    return '- Please ensure device is raised from surface to allow rear ' \
           'camera to take a photograph during camera tests.'


def print_warnings():
    print_separator('*')
    logger.warning(get_camera_warning())
    print_separator('*')


def print_separator(char):
    logger.warning(char * 125)


def clean_output_directory(config_stack):
    """
    Clean the output directory specified in config_stack.

    This is required so that adt-run does not fail due to existing content
    in the output directory from a previous test run.

    """
    path = os.path.abspath(config_stack.get('output_dir'))
    if os.path.isdir(path):
        clean_dir(path)


def _get_config_stack(conf_name, config_section, silent=False):
    """Get the configuration values for the system tests.

    This function registers the declared options, opens the config store
    using the specified file name, adds the options coming from the command
    line, and asks the user for the missing options.

    :param conf_name: File name of config file to use.
    :param config_section: Section of config to use.
    :param silent: If True then don't prompt the user for any missing config
                   options, use default values instead. If False then prompt
                   the user for any missing option values.

    """
    config_stack = config.UbuntuSystemTestsConfig(
        conf_name, section=config_section)
    config_stack.get_config_from_file()
    config_stack.get_missing_options(silent)
    return config_stack


def enable_autopilot_introspection(config_stack):
    """Enable autopilot introspection on the required device."""
    command = cmds.build_enable_autopilot_introspection_command(config_stack)
    subprocess.check_output(command)


def run_autopkg_setup(args):
    """Run setup commands on the required device, rebooting if required.
    :param args: Command line arguments for setup command.
    :return: Setup status code, 0 indicating setup was successful. Any error
             status encountered during setup will be translated to error
             status 16 from autopkgtest which indicates a testbed failure.
    """
    config_stack = _prepare_config(args)
    adt_run_cmd = cmds.build_adt_run_setup_command(
        config_stack,
        args.config_section,
        cache_deps=args.cache_deps,
        dist_upgrade=args.dist_upgrade,
        install_tests=args.install_tests,
        network_ready=args.network_ready,
        update_apt=args.update_apt,
        verbose=args.verbose)

    return run_setup_commands(adt_run_cmd, config_stack)


def run_setup_commands(setup_cmd, config_stack):
    """Run setup commands to prepare the device for test execution.

    This will run the setup command on the device. If this causes a failure
    on the device, or a reboot is required, the device will reboot. The setup
    command will then be executed on the device again. If the second attempt
    at setup fails then the failure status will be reported and setup will
    stop.

    :param setup_cmd: Command to execute to set the device up.
    :param config_stack: Device configuration data.
    :return: Setup status code, 0 indicating setup was successful. Any error
             status encountered during setup will be translated to error
             status 16 from autopkgtest which indicates a testbed failure.
    """
    count = 0
    status = 1
    while (status and count < 2):
        clean_output_directory(config_stack)
        status = subprocess.call(setup_cmd)
        if status and count < 1:
            # Error was reported after first setup attempt,
            # so reboot device and try again.
            logger.warning(
                'Rebooting device after setup completed with '
                'status {s}.'.format(s=status))
            reboot_status = _reboot_device(config_stack)
            if reboot_status:
                logger.warning(
                    'Device reboot completed with status '
                    '{s}.'.format(s=reboot_status))
                return reboot_status
        count += 1
    return status


def _reboot_device(config_stack):
    """Reboot the device specified in config.
    :param config_stack: Device configuration.
    :return: Status of reboot command.
    """
    for reboot_command in cmds._get_adb_reboot_commands(config_stack):
        status = subprocess.call(reboot_command)
        if status:
            # Error status reported from reboot, so don't attempt any further
            # commands and return status value.
            break
    return status


def run_autopkg_test(config_stack, args):
    """ Run the autopkgtest in the device for the tests selected
    :param config_stack: The config objects
    :param args: command line arguments
    :return: the autopkgtest execution exit code
    """
    extra_depends = debian.get_extra_dependencies(args.suites)
    adt_run_cmd = cmds.build_adt_run_test_command(
        config_stack,
        args.config_section,
        tests_to_run=args.suites,
        extra_depends=extra_depends,
        network_ready=args.network_ready,
        touch_visualization=args.touch_visualization,
        verbose=args.verbose)
    return subprocess.call(adt_run_cmd)


def generate_user_report(config_stack):
    """Create a human readable report from the subunit results."""
    artifacts_directory = results.get_artifacts_directory(config_stack)
    results.convert_subunit_results_to_pyunit(artifacts_directory)
    results.print_results(artifacts_directory)


if __name__ == '__main__':
    sys.exit(main())
