#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from ubuntu_system_tests.common import (
    config,
    ssh,
)
from ubuntu_system_tests.host import (
    debian,
    testdesc,
)

# apt-cacher-ng definitions
APT_CACHER = 'apt-cacher-ng'
APT_CACHER_CACHE_DIR = '/var/cache/apt-cacher-ng'
APT_CACHER_LOG_DIR = '/var/log/apt-cacher-ng/'
APT_CACHER_INIT_FILE = '/etc/init/apt-cacher-ng.conf'
APT_PROXY = 'Acquire::http { Proxy "http://127.0.0.1:3142"; };'
APT_PROXY_FILE = '/etc/apt/apt.conf.d/01proxy'
EXT_CACHE_DIR = '/userdata/apt-cacher-ng/cache'
EXT_LOG_DIR = '/userdata/apt-cacher-ng/logs'
PERMISSION = '755'

QTTESTABILITY_AUTOPILOT = 'qttestability-autopilot'

PPA_LIST = ['ppa:canonical-platform-qa/selenium',
            'ppa:canonical-platform-qa/ubuntu-system-tests']
INSTALL_LIST = [QTTESTABILITY_AUTOPILOT]


_ADT_RUN_SHORT_TIMEOUT = 600  # 10 minutes.
_ADT_RUN_TEST_TIMEOUT = 86400  # 24 hours

REBOOT_FILE = '/tmp/ust-reboot-required'

STATUS_NO_CONNECTION = 1
STATUS_FS_RO_MOUNT_ERROR = 2
STATUS_FS_RW_MOUNT_ERROR = 3
STATUS_REBOOT_REQUIRED = 4


def build_basic_adt_run_test_command(config_stack, extra_depends, verbose):
    """Build and return first part of adt-run command for test execution."""
    run_cmd = _get_adt_run_cmd_start(config_stack, verbose)
    if extra_depends:
        run_cmd.append('--override-control={}'.format(
            debian.create_temp_control_file(extra_depends))
        )
    run_cmd.extend([
        # Updating and installing can take a while.
        '--timeout-short={}'.format(_ADT_RUN_SHORT_TIMEOUT),
        # Regression tests can take up to 8 hours, no reason to cut short.
        '--timeout-test={}'.format(_ADT_RUN_TEST_TIMEOUT),
        '--testname=systemtests',
        '-B',
        '--unbuilt-tree={}'.format(debian.get_debian_path()),
        # Even if the config file on the host has a different name, on the
        # testbed it's copied with the default name.
        '--copy',
        '{src}:{dest}'.format(
            src=config_stack.file_path,
            dest=config.get_device_config_file_path()),
    ])
    return run_cmd


def build_basic_adt_run_setup_command(config_stack, verbose, cache_deps):
    """Build and return first part of adt-run command for device setup."""
    run_cmd = _get_adt_run_cmd_start(config_stack, verbose)
    if cache_deps:
        # Run the setup script from the control file listing all required
        # dependencies. This will ensure all required dependencies are
        # downloaded and cached as part of the setup command.
        run_cmd.append('--testname=setup')
    else:
        # Dependency caching is not required. In this case don't download
        # any dependencies by using a control file with no dependencies.
        tmp_control = debian.create_temp_file('Tests: setup\nDepends: \n')
        run_cmd.append('--override-control={}'.format(tmp_control))
    run_cmd.extend([
        '-B',
        '--unbuilt-tree={}'.format(debian.get_debian_path())])
    return run_cmd


def build_adt_run_setup_command(config_stack, config_section, update_apt=False,
                                network_ready=False, dist_upgrade=False,
                                install_tests=False, cache_deps=False,
                                verbose=False):
    """Build and return adt-run command for device setup."""
    run_cmd = build_basic_adt_run_setup_command(
        config_stack, verbose, cache_deps)

    run_cmd.extend(_get_delete_reboot_required_setup_command())
    run_cmd.extend(_get_set_config_section_command(config_section))
    run_cmd.extend(_get_network_setup_command(config_stack, network_ready))
    run_cmd.extend(_get_disable_auto_reboot_setup_command())
    run_cmd.extend(_get_add_ppa_if_required_setup_command())
    run_cmd.extend(_get_update_apt_setup_command(update_apt))
    run_cmd.extend(_get_install_test_pkgs_if_required_setup_command())
    run_cmd.extend(_get_apt_cacher_install_setup_command(cache_deps))
    run_cmd.extend(_get_dist_upgrade_if_required_setup_command(dist_upgrade))
    run_cmd.extend(_get_install_tests_setup_command(install_tests))
    run_cmd.extend(_get_fs_ro_setup_command())
    run_cmd.extend(_get_apparmor_rules_update_setup_command())
    run_cmd.extend(_get_check_reboot_required_setup_command())

    return run_cmd + ['---'] + _get_ssh_command(config_stack)


def build_adt_run_test_command(config_stack, config_section,
                               tests_to_run=None, extra_depends=None,
                               network_ready=False, touch_visualization=False,
                               verbose=False):
    """Build and return adt-run command for executing tests on device."""
    run_cmd = build_basic_adt_run_test_command(
        config_stack, extra_depends, verbose)

    run_cmd.extend(_get_fs_ro_setup_command())
    run_cmd.extend(_get_set_tests_to_run_command(tests_to_run))
    run_cmd.extend(_get_set_config_section_command(config_section))
    run_cmd.extend(_get_set_touch_visualization(touch_visualization))
    run_cmd.extend(_get_network_setup_command(config_stack, network_ready))
    run_cmd.extend(_get_apparmor_rules_update_setup_command())
    run_cmd.extend(ssh.get_ssh_config_commands())

    return run_cmd + ['---'] + _get_ssh_command(config_stack)


def _get_adt_run_cmd_start(config_stack, verbose):
    """Return the start of adt_run command."""
    output_dir = config_stack.get('output_dir')
    run_cmd = ['adt-run', '--output-dir', output_dir]
    if verbose:
        run_cmd.append('-ddd')
    return run_cmd


def _get_ssh_command(config_stack):
    """Return command parameters for ssh connection to requried device."""
    device_password = config_stack.get('device_password')
    return (['ssh', '-s', 'adb', '--', '-p', device_password] +
            _get_serial_argument(config_stack))


def build_enable_autopilot_introspection_command(config_stack):
    """Return command to enable autopilot introspection on required device."""
    command = (
        ['phablet-config'] +
        _get_serial_argument(config_stack) +
        ['autopilot', '--dbus-probe', 'enable']
    )
    return command


def _get_adb_reboot_commands(config_stack):
    """Return list of adb commands to reboot device."""
    return [
        _get_adb_command(config_stack, 'shell', 'sync'),
        _get_adb_command(config_stack, 'reboot')]


def _get_adb_command(config_stack, *args):
    """Return adb command to run specified command args on connected device."""
    return ['adb'] + _get_serial_argument(config_stack) + list(args)


def _get_serial_argument(config_stack):
    """Return serial commands from config, or empty list if not specified."""
    device_serial = config_stack.get('device_serial')
    if device_serial:
        return ['-s', device_serial]
    else:
        # No serial in the config, run the commands as if only one device is
        # connected.
        return []


def _get_set_tests_to_run_command(tests_to_run):
    """Return command to set list of tests as environment variable."""
    tests = ' '.join(tests_to_run or [])
    return _get_set_env_var_adt_command('TESTS_TO_RUN', tests)


def _get_set_config_section_command(config_section):
    """Return command to set config section name as environment variable."""
    return _get_set_env_var_adt_command('CONFIG_SECTION', config_section)


def _get_set_touch_visualization(touch_visualization):
    """Return command to set touch visualization as environment variable."""
    return _get_set_env_var_adt_command(
        'TOUCH_VISUALIZATION', str(touch_visualization).lower())


def _get_set_env_var_adt_command(name, value):
    """Return command to set environment variable on device."""
    return ['--env={n}={v}'.format(n=name, v=value)]


def _get_update_apt_command():
    """Return command to apt-get update."""
    return ('echo "apt-get updating"; apt-get update || '
            '(sleep 15; apt-get update); ')


def _get_update_apt_setup_command(update_apt):
    """Return setup command to apt-get update."""
    if update_apt:
        return [
            '--setup-commands',
            _get_fs_rw_command() +
            _get_update_apt_command()]
    return []


def _get_device_network_command(wifi_ssid, wifi_password):
    """Return command to configure wifi network on device.

    :param wifi_ssid: SSID of required wifi network
    :param wifi_password: Password of required wifi network

    """
    cmd_fmt = ('nmcli device wifi connect '
               '\'{ssid}\' password \'{password}\' 2>/dev/null || true')
    return cmd_fmt.format(ssid=wifi_ssid, password=wifi_password)


def _get_is_fs_ro_command():
    """Return command to indicate if file system is currently read-only."""
    return 'grep \' /\ ext\' /proc/mounts | grep -q \'ro\''


def _get_is_fs_rw_command():
    """Return command to indicate if file system is currently writable."""
    return 'grep \' /\ ext\' /proc/mounts | grep -q \'rw\''


def _get_fs_rw_command():
    """Return command to set file system read-write if not already."""
    return _get_if_command(_get_is_fs_ro_command(), _get_set_fs_rw_command())


def _get_fs_ro_setup_command():
    """Return setup command to set file system read-only if not already."""
    return [
        '--setup-commands',
        _get_if_command(_get_is_fs_rw_command(), _get_set_fs_ro_command())
        ]


def _get_set_fs_rw_command():
    """Return command to set file system writable, retyring on failure."""
    return ('echo "Mounting file system rw"; '
            'sync; RETRY=0; until mount -o remount,rw / || [ $RETRY -gt 10 ]; '
            'do sleep 1; RETRY=$(($RETRY+1)); echo "mount attempt failed, '
            'trying again..."; done; if [ $RETRY -gt 10 ]; then exit '
            '{status}; fi; '.format(status=STATUS_FS_RW_MOUNT_ERROR))


def _get_set_fs_ro_command():
    """Return command to set file system read-only, retyring on failure."""
    return ('echo "Mounting file system ro"; '
            'sync; RETRY=0; until mount -o remount,ro / || [ $RETRY -gt 10 ]; '
            'do sleep 1; RETRY=$(($RETRY+1)); echo "mount attempt failed, '
            'trying again..."; done; if [ $RETRY -gt 10 ]; then exit '
            '{status}; fi; '.format(status=STATUS_FS_RO_MOUNT_ERROR))


def _get_network_setup_command(config_stack, network_ready):
    """ The network needs to be configured on the device
    This setup-command should be the first to run so that wi-fi is ready
    """
    cmd = []
    if not network_ready:
        network_cmd = _get_device_network_command(
            config_stack.get('wifi_ssid'),
            config_stack.get('wifi_password'))
        cmd = ['--setup-commands', network_cmd]

    return cmd + [
        '--setup-commands', 'echo \'Checking device network connection...\'; '
        'RETRY=0; until wget -q --spider launchpad.net || [ $RETRY -gt 60 ]; '
        'do sleep 1; RETRY=$(($RETRY+1)); echo "wget attempt $RETRY failed, '
        ' trying again..."; done; if [ $RETRY -gt 60 ]; then echo \'Device '
        'has no network connection!\'; exit {status}; fi'.format(
            status=STATUS_NO_CONNECTION)
    ]


def _get_apt_cacher_autostart_command():
    """Write the apt-cacher-ng conf file into /etc/init/apt-cacher-ng.conf
    to re-launch apt-cacher-ng on each subsequent reboot.

    This method assumes that upstart is used as the init system. If this
    changes in future then this method will need updating to accomodate that.
    """
    script = ('start on runlevel [2345]\n'
              'script\n' +
              _get_mount_dir_command(EXT_LOG_DIR, APT_CACHER_LOG_DIR) +
              _get_mount_dir_command(EXT_CACHE_DIR, APT_CACHER_CACHE_DIR) +
              _get_apt_cacher_start_command() +
              '\nend script')
    return _get_create_file_command(content=script, path=APT_CACHER_INIT_FILE)


def _get_apt_cacher_install_setup_command(apt_cacher_required):
    """Return command to install apt-cacher-ng on device and configure it to
    use userdata partion for caching and logging. If apt-cacher-ng is already
    installed on the device then no action will be taken."""
    if not apt_cacher_required:
        return []
    is_apt_cacher_installed = _get_packages_installed_command(APT_CACHER)
    install_apt_cacher_cmds = (
        _get_fs_rw_command() +
        _get_install_packages_command(APT_CACHER) +
        _get_apt_cacher_stop_command() +
        _get_create_folder_command(
            path=EXT_LOG_DIR, owner=APT_CACHER, permissions=PERMISSION) +
        _get_create_folder_command(
            path=EXT_CACHE_DIR, owner=APT_CACHER, permissions=PERMISSION) +
        _get_create_folder_command(
            path=APT_CACHER_CACHE_DIR, owner=APT_CACHER,
            permissions=PERMISSION) +
        _get_mount_dir_command(EXT_LOG_DIR, APT_CACHER_LOG_DIR) +
        _get_mount_dir_command(EXT_CACHE_DIR, APT_CACHER_CACHE_DIR) +
        _get_create_file_command(content=APT_PROXY, path=APT_PROXY_FILE) +
        _get_apt_cacher_start_command() +
        _get_apt_cacher_autostart_command())
    return ['--setup-commands', _get_if_not_command(
        is_apt_cacher_installed, install_apt_cacher_cmds)]


def _get_if_not_command(condition, commands):
    """Return if command for specified condition and commands."""
    return 'if ! ({condition}); then {commands} fi; '.format(
        condition=condition, commands=commands)


def _get_if_command(condition, commands):
    """Return if not command for specified condition and commands."""
    return 'if ({condition}); then {commands} fi; '.format(
        condition=condition, commands=commands)


def _get_create_folder_command(path, owner, permissions):
    """Return command to create a folder with specified path, owner and
    permissions."""
    return ('if ! [ -d {path} ]; then '
            'mkdir -p {path}; fi; '
            'chown {owner}:{owner} {path}; '
            'chmod {permissions} {path}; ').format(
        path=path, owner=owner, permissions=permissions)


def _get_install_test_packages_command():
    """Return command to install required test packages."""
    return (_get_fs_rw_command() +
            _get_install_packages_command(' '.join(INSTALL_LIST)))


def _get_test_packages_installed_command():
    """Return command to indicate whether all required packages are already
    installed or not.
    :param apt_cacher_required: Whether apt-cahcer-ng is required to be
                                installed or not.
    """
    return _get_packages_installed_command(' '.join(INSTALL_LIST))


def _get_packages_installed_command(packages):
    """Return command to tell if all listed packages are installed or not."""
    return ('! apt-cache policy {packages} | '
            'grep -q \'Installed: (none)\'; '.format(packages=packages))


def _get_add_ppa_required_command():
    """Return command to indicate if any ppa is required to be added."""
    return ('! ls /etc/apt/sources.list.d/canonical-platform-qa* > /dev/null '
            '2>&1')


def _get_add_ppas_command():
    """Return command to add all required ppas."""
    cmd = ''
    for ppa in PPA_LIST:
        cmd += 'apt-add-repository -y {ppa}; '.format(ppa=ppa)
    return _get_fs_rw_command() + cmd


def _get_add_ppa_if_required_setup_command():
    """Return setup command to add required ppas to device if required."""
    return ['--setup-commands', _get_if_command(
        _get_add_ppa_required_command(), _get_add_ppas_command())]


def _dist_upgrade_required_command():
    """Return command to indicate if any packages need to be updated as part
    of a dist-upgrade command."""
    return ('su - $ADT_NORMAL_USER -c \'apt-get dist-upgrade -s\' | '
            'grep -q \'Inst\\|Conf\\|Remv\'; ')


def _get_dist_upgrade_if_required_setup_command(required):
    """Return setup command to perform dist-upgrade if required."""
    if not required:
        return []
    return ['--setup-commands', _get_if_command(
        _dist_upgrade_required_command(), _get_dist_upgrade_command())]


def _get_install_test_pkgs_if_required_setup_command():
    """Return setup command to install test packages if required."""
    return ['--setup-commands',
            _get_if_not_command(
                _get_test_packages_installed_command(),
                _get_install_test_packages_command())]


def _get_install_packages_command(packages):
    """Return command to install packages on device."""
    return ('apt-get -y --no-install-recommends install {packages}; ').format(
        packages=packages)


def _get_mount_dir_command(target, source):
    """Return command to create a symlink between source and target
    directories."""
    return 'mount -o bind {target} {source}; '.format(
        target=target, source=source)


def _get_create_file_command(content, path):
    """Return command to create file with specified content and path."""
    return 'echo \'{content}\' | tee {path}; '.format(
        content=content, path=path)


def _get_apt_cacher_start_command():
    """Return command to start apt-cacher-ng process."""
    return ('/usr/sbin/apt-cacher-ng -c /etc/apt-cacher-ng '
            'pidfile=/var/run/apt-cacher-ng/pid '
            'SocketPath=/var/run/apt-cacher-ng/socket '
            'foreground=0; ')


def _get_apt_cacher_stop_command():
    """Return command to stop apt-cacher-ng process."""
    return 'pkill apt-cacher-ng || true; '


def _get_dist_upgrade_command():
    """Return command to dist-upgrade device."""
    return (_get_set_reboot_required_command() + 'apt-get -y dist-upgrade; ')


def _get_check_reboot_required_setup_command():
    """If a reboot is required, then exit with status to indicate that."""
    return [
        '--setup-commands',
        'if [ -f {p} ]; then exit {s}; fi; '.format(
            p=REBOOT_FILE, s=STATUS_REBOOT_REQUIRED)]


def _get_set_reboot_required_command():
    """Create file to indicate that reboot is required."""
    return 'if [ ! -f {p} ]; then touch {p}; fi; '.format(p=REBOOT_FILE)


def _get_delete_reboot_required_setup_command():
    """Delete file to indicate that reboot is required."""
    return [
        '--setup-commands',
        'if [ -f {p} ]; then rm -f {p}; fi; '.format(p=REBOOT_FILE)]


def _get_disable_auto_reboot_setup_command():
    """Return setup command to disable automatic rebooting by autopkgtest."""
    return [
        '--setup-commands',
        'if [ ! -f /run/autopkgtest_no_reboot.stamp ]; then '
        'touch /run/autopkgtest_no_reboot.stamp; fi; ']


def _get_install_tests_setup_command(required):
    """ Get a list with all the dependencies from the tests/control file and
    when the apt-get install does not install any package, the device is set
    as ro and exit, otherwise it is rebooted and will start as ro
    """
    if not required:
        return []
    dir_name = debian.get_debian_path()
    dependencies = testdesc.get_all_dependencies(dir_name, 'TMPINSTALL')
    install_command = 'apt-get -o dir::cache::archives="/tmp/" -y install'
    clean_command = 'apt-get clean; '
    for dep in dependencies:
        install_command += ' {}'.format(dep.split()[0])
    install_command += '; sleep 5; '

    return ['--setup-commands',
            _get_fs_rw_command() +
            clean_command +
            install_command +
            clean_command]


def _get_apparmor_rules_update_setup_command():
    """Return setup command to update apparmor for autopilot introspection."""
    rules_file = '/var/cache/apparmor/click-ap.rules'
    update_command = 'printf "dbus (receive, send) ' \
                     'bus=session path=/com/canonical/Autopilot/**,\n' \
                     '/tmp/adt-run.** r,\n" > {rules};' \
                     'aa-clickhook --include={rules}'.format(rules=rules_file)
    return [
        '--setup-commands',
        'if [ ! -e {rules} ];then {command};fi'.format(rules=rules_file,
                                                       command=update_command)
    ]
