#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import deb822
import glob
import json
import os
import tempfile

# Installation path for shared files
INSTALLED_SHARE_DIR = '/usr/share/ubuntu-system-tests'
# Installation path for installed python scripts
INSTALLED_PYTHON_DIR = '/usr/lib/python'


def get_extra_dependencies(tests_to_run):
    """
    Get any extra dependencies required to run the list of
    tests provided

    :param tests_to_run: The list of tests to run
    :returns: A list of the extra dependencies
    """
    depends_map = _read_dependency_files(
        os.path.realpath('ubuntu_system_tests/external_tests')
    )
    extra_depends = []
    for test in tests_to_run:
        extra_depends.extend(_extra_dependencies_for_test(test, depends_map))
    return list(set(extra_depends))


def _read_dependency_files(deps_path):
    """
    Read the information about how dependencies map
    to test names from the .json files in the path
    provided.

    :param deps_path: The path to the directory containing the files
    :returns: A dictionary associating test prefixes with lists of dependencies
    """
    depends_map = {}
    for file in glob.glob(deps_path + '/*.json'):
        with open(file) as deps_file:
            for depends in json.loads(deps_file.read())['test_dependencies']:
                depends_map[depends['id']] = depends['depends']
    return depends_map


def _extra_dependencies_for_test(test, depends_map):
    """
    Calculates the extra dependencies required for a particular test

    :param test: The test to get the dependencies for
    :param depends_map: The dependency mapping
    :returns: A list of the extra dependencies
    """
    extra_dependencies = []
    for id in depends_map.keys():
        if test.startswith(id):
            extra_dependencies.append(depends_map[id])
    return extra_dependencies


def create_temp_control_file(dependencies,
                             control_file='debian/tests/control'):
    """Create a copy of the test suite control file, appending the list of
    dependencies

    :param dependencies: List of dependencies to append to file
    :returns: File path to the new control file
    """
    # open existing control file
    with open(os.path.realpath(control_file), 'r') as tests_control:
        new_content = _append_dependencies(
            tests_control, dependencies)
    return create_temp_file(new_content)


def create_temp_file(content):
    """Create a temporary file with specified content and return path to it."""
    # create a temporary file and write content to it
    _, tmp_path = tempfile.mkstemp(text=True)
    with open(tmp_path, 'w') as tmp_file:
        tmp_file.write(content)
    return tmp_path


def _append_dependencies(tests_control, dependencies):
    """Append the list of dependencies to the file content and return"""
    content = ''
    for pkg in deb822.Packages.iter_paragraphs(tests_control):
        if pkg['Tests'].startswith('systemtests'):
            current_deps = pkg['Depends']
            for dep in dependencies:
                current_deps += '\n         {},'.format(dep)
            pkg['Depends'] = current_deps
        content += pkg.dump() + '\n'
    return content.rstrip()


def _is_installed():
    """Return True if the suite is running from installed location."""
    return os.path.realpath(__file__).startswith(INSTALLED_PYTHON_DIR)


def get_debian_path():
    """Return the path of the debian folder.
    This will depend on whether tests are running from installed location
    or local source directory.
    """
    return INSTALLED_SHARE_DIR if _is_installed() else os.getcwd()
