# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from unittest import mock

import testscenarios

from ubuntu_system_tests import selftests
from ubuntu_system_tests.host import command_line
from ubuntu_system_tests.common import config
from ubuntu_system_tests.selftests.utils import (
    get_cmdline_run_params,
    get_cmdline_setup_params,
)


class SetupSystemTestsCommandsTestCase(selftests.ConfigBaseTestCase,
                                       testscenarios.TestWithScenarios):

    scenarios = [
        # The nmcli network setup command is passed as --setup-command option
        # to adt-run. Verify that it should not run as separate command.
        ('without network setup', {
            'keyword_arguments': {'update_apt': False,
                                  'network_ready': True},
            'expected_commands': ['adt-run'],
        }),
        ('with network setup', {
            'keyword_arguments': {'update_apt': False,
                                  'network_ready': False},
            'expected_commands': ['adt-run'],
        }),
    ]

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    @mock.patch('ubuntu_system_tests.host.command_line.generate_user_report')
    def test_setup_system_tests_commands(self, mock_generate_user_report):
        def get_command_from_call(call):
            """Return the name of the command, without the arguments."""
            cmd_list = call[0][0]
            if type(cmd_list) == str:
                cmd_list = cmd_list.split(' ')
            return cmd_list[0]

        self.set_config_values()
        conf_name = os.path.basename(self.config_stack.file_path)
        params = get_cmdline_setup_params(
            config=conf_name, **self.keyword_arguments)

        # Do not execute the subprocesses, just record their calls.
        with mock.patch('subprocess.Popen') as mock_popen:
            enter = mock_popen().__enter__()
            # Make sure that check_output and call return with no errors.
            enter.communicate = mock.Mock(return_value=(0, 'dummy'))
            enter.poll = mock.Mock(return_value=0)
            enter.wait = mock.Mock(return_value=0)
            command_line.setup_system_tests(params)

        # The first call doesn't run any command.
        calls = mock_popen.call_args_list[1:]
        commands = [get_command_from_call(call) for call in calls]
        self.assertEqual(self.expected_commands, commands)


class SetupCommandLineArgumentsTestCase(testscenarios.TestWithScenarios):

    SETUP_CMD = 'setup'
    DEFAULT_CONFIG_FILE = 'ubuntu-system-tests.conf'
    DEFAULT_CONFIG_SECT = config.KEY_DEFAULT
    DO_APT_UPDATE = True
    DONT_APT_UPDATE = False
    NETWORK_ALREADY_SETUP = True
    NETWORK_NOT_ALREADY_SETUP = False
    DO_DIST_UPGRADE = True
    DONT_DIST_UPGRADE = False
    DO_INSTALL_TESTS = True
    DONT_INSTALL_TESTS = False
    DO_USE_CACHE_DEPS = True
    DONT_USE_CACHE_DEPS = False
    DO_USE_SILENT_MODE = True
    DONT_USE_SILENT_MODE = False
    DO_USE_VERBOSE_LOGGING = True
    DONT_USE_VERBOSE_LOGGING = False

    scenarios = [
        ('default setup arguments', {
            'command_line_arguments': [],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('alternative config', {
            'command_line_arguments': ['--config=alternative.conf'],
            'parsed_arguments': (
                SETUP_CMD,
                'alternative.conf',
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('alternative config section', {
            'command_line_arguments': ['--config-section=custom-section'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                'custom-section',
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('only apt-get update', {
            'command_line_arguments': ['--update-apt'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_CACHE_DEPS,
                DONT_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('with network already setup', {
            'command_line_arguments': ['--network-ready'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('with dist-upgrade enabled', {
            'command_line_arguments': ['--dist-upgrade'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('with install dependencies enabled', {
            'command_line_arguments': ['--install-tests'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_CACHE_DEPS,
                DONT_DIST_UPGRADE,
                DO_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('with cache deps specified', {
            'command_line_arguments': ['--cache-deps'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DONT_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
        ('with verbose logging specified', {
            'command_line_arguments': ['--verbose'],
            'parsed_arguments': (
                SETUP_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DO_USE_VERBOSE_LOGGING,
                DO_USE_CACHE_DEPS,
                DO_DIST_UPGRADE,
                DONT_INSTALL_TESTS,
                DO_APT_UPDATE,
                DONT_USE_SILENT_MODE,
            )
        }),
    ]

    @mock.patch('ubuntu_system_tests.host.command_line.run_autopkg_setup')
    def test_main_must_pass_command_line_arguments_to_run_system_tests(
            self, mock_run_autopkg_setup):
        command_line.main([self.SETUP_CMD] + self.command_line_arguments)
        params = get_cmdline_setup_params(*self.parsed_arguments)
        mock_run_autopkg_setup.assert_called_once_with(params)


class RunCommandLineArgumentsTestCase(testscenarios.TestWithScenarios):

    RUN_CMD = 'run'
    DEFAULT_CONFIG_FILE = 'ubuntu-system-tests.conf'
    DEFAULT_CONFIG_SECT = config.KEY_DEFAULT
    ALL_SUITES = []
    NETWORK_ALREADY_SETUP = True
    NETWORK_NOT_ALREADY_SETUP = False
    DO_USE_VERBOSE_LOGGING = True
    DONT_USE_VERBOSE_LOGGING = False
    DO_USE_TOUCH_VISUALIZATION = True
    DONT_USE_TOUCH_VISUALIZATION = False
    DO_USE_SILENT_MODE = True
    DONT_USE_SILENT_MODE = False

    scenarios = [
        ('default run arguments', {
            'command_line_arguments': [],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('alternative config', {
            'command_line_arguments': ['--config=alternative.conf'],
            'parsed_arguments': (
                RUN_CMD,
                'alternative.conf',
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('alternative config section', {
            'command_line_arguments': ['--config-section=custom-section'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                'custom-section',
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('with network already setup', {
            'command_line_arguments': ['--network-ready'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('with verbose logging specified', {
            'command_line_arguments': ['--verbose'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DO_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ALL_SUITES,
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('single test', {
            'command_line_arguments': ['suite.tests.test_1'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ['suite.tests.test_1'],
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
        ('mutliple tests', {
            'command_line_arguments': [
                'suite.tests.test_1', 'suite.tests.test_2'],
            'parsed_arguments': (
                RUN_CMD,
                DEFAULT_CONFIG_FILE,
                DEFAULT_CONFIG_SECT,
                NETWORK_NOT_ALREADY_SETUP,
                DONT_USE_VERBOSE_LOGGING,
                DONT_USE_SILENT_MODE,
                ['suite.tests.test_1', 'suite.tests.test_2'],
                DONT_USE_TOUCH_VISUALIZATION,
            )
        }),
    ]

    @mock.patch('ubuntu_system_tests.host.command_line.run_system_tests')
    def test_main_must_pass_command_line_arguments_to_run_system_tests(
            self, mock_run_system_tests):
        command_line.main([self.RUN_CMD] + self.command_line_arguments)
        params = get_cmdline_run_params(*self.parsed_arguments)
        mock_run_system_tests.assert_called_once_with(params)
