# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import mock

from ubuntu_system_tests import selftests
from ubuntu_system_tests.host import commands
from ubuntu_system_tests.common import config


def check_any_item_in_list_contains_value(list, value):
    return any(item.find(value) >= 0 for item in list)


def check_any_item_in_list_startswith_value(list, value):
    return any(item.startswith(value) for item in list)


class RunCommandTestCase(selftests.ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_setup_network_command_contains_credentials(self):
        self.set_config_values()

        command = commands._get_device_network_command(
            'test_ssid', 'test_password')

        self.assertTrue(command.startswith(
            'nmcli device wifi connect \'test_ssid\' '
            'password \'test_password\''))

    def test_enable_introspection_with_serial(self):
        serial = 'test_serial'
        self.set_config_values(device_serial=serial)

        command = commands.build_enable_autopilot_introspection_command(
            self.config_stack)

        self.assertEqual(['phablet-config', '-s', 'test_serial'], command[:3])

    def test_enable_introspection_without_serial(self):
        self.set_config_values(device_serial='')

        command = commands.build_enable_autopilot_introspection_command(
            self.config_stack)

        self.assertNotIn('-s', command)

    def test_adt_run_must_be_called_with_short_timeout(self):
        self.set_config_values()

        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT)
        command = ' '.join(command)

        run_arguments, _ = command.split('---')
        self.assertIn('--timeout-short=600', run_arguments)

    def test_adt_run_command_no_rw_file_mount(self):
        self.set_config_values()
        run_cmd = ' '.join(commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT))
        self.assertNotIn('mount -o remount,rw /', run_cmd)

    def test_adt_run_command_ro_file_mount(self):
        self.set_config_values()
        run_cmd = ' '.join(commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT))
        self.assertIn('mount -o remount,ro /', run_cmd)

    def test_adt_run_with_serial(self):
        serial = 'test_serial'
        self.set_config_values(device_serial=serial)

        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT)
        command = ' '.join(command)

        _, adb_arguments = command.split('ssh -s adb --')
        self.assertIn('-s test_serial', adb_arguments)

    def test_adt_run_without_serial(self):
        self.set_config_values(device_serial='')

        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT)
        command = ' '.join(command)

        _, adb_arguments = command.split('ssh -s adb --')
        self.assertNotIn('-s test_serial', adb_arguments)

    def test_adt_run_command_with_network_ready_creates_no_network(self):
        self.set_config_values()

        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT, network_ready=True)

        self.assertFalse(check_any_item_in_list_startswith_value(
            command, 'nmcli device wifi connect'))

    def test_adt_run_command_with_default_parameters_creates_network(self):
        self.set_config_values()

        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT)

        self.assertTrue(check_any_item_in_list_startswith_value(
            command, 'nmcli device wifi connect'))

    def test_enabling_verbose_option(self):
        self.set_config_values()
        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT, verbose=True)
        self.assertTrue(check_any_item_in_list_contains_value(command, '-ddd'))

    def test_verbose_disabled_by_default(self):
        self.set_config_values()
        command = commands.build_adt_run_test_command(
            self.config_stack, config.KEY_DEFAULT)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, '-ddd'))


class SetupCommandTestCase(selftests.ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_ppa_setup_commands_set_image_readwrite(self):
        self.set_config_values()
        self.assertIn(
            'mount -o remount,rw /',
            ' '.join(commands.build_adt_run_setup_command(
                self.config_stack, config.KEY_DEFAULT))
        )

    def test_ppa_setup_commands_used_in_adt_setup(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT)
        command = ' '.join(command)

        run_arguments, _ = command.split('---')
        self.assertIn(
            'apt-add-repository -y ppa:canonical-platform-qa',
            run_arguments
        )

    def test_setup_commands_install_testability_libs(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command,
                'apt-get -y --no-install-recommends install '
                'qttestability-autopilot')
        )

    def test_setting_fs_rw_is_conditional_on_current_state(self):
        self.assertIn(
            commands._get_is_fs_ro_command(), commands._get_fs_rw_command())

    def test_setting_fs_ro_is_conditional_on_current_state(self):
        command = commands._get_fs_ro_setup_command()
        check_any_item_in_list_contains_value(
            command, commands._get_is_fs_rw_command())

    def test_enabling_verbose_option(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT, verbose=True)
        self.assertTrue(check_any_item_in_list_contains_value(command, '-ddd'))

    def test_verbose_disabled_by_default(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, '-ddd'))

    def test_cache_deps_uses_default_control_file(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT, cache_deps=True)
        self.assertFalse(
            check_any_item_in_list_contains_value(
                command, '--override-control'))

    def test_no_cache_deps_overrides_control_file(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command, '--override-control'))

    def test_cache_deps_installs_apt_cacher(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT, cache_deps=True)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command,
                'apt-get -y --no-install-recommends install apt-cacher-ng'))

    def test_default_no_apt_cacher_install(self):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, 'apt-cacher-ng'))

    @mock.patch('ubuntu_system_tests.host.testdesc.get_all_dependencies')
    def test_install_tests_adds_extra_dependencies(
            self, mock_get_all_dependencies):
        self.set_config_values()
        mock_get_all_dependencies.return_value = ['test-dep-1', 'test-dep-2']
        command = commands.build_adt_run_setup_command(
            self.config_stack, config.KEY_DEFAULT, install_tests=True)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command, 'install test-dep-1 test-dep-2'))
