# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import fixtures
import json
import os
from tempfile import NamedTemporaryFile
from unittest import mock

from ubuntu_system_tests import selftests
from ubuntu_system_tests.host import debian
from ubuntu_system_tests.common import config


def check_any_item_in_list_contains_value(list, value):
    return any(item.find(value) for item in list)


def check_any_item_in_list_startswith_value(list, value):
    return any(item.startswith(value) for item in list)


class CommandTestCase(selftests.ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_get_extra_dependencies(self):
        tests = ('app1.tests.test_module.TestClass.test_x '
                 'app2.tests.test_module.TestClass.test_y '
                 'app3.tests.test_module.TestClass.test_z')
        self.set_config_values(tests_to_run=tests)
        expected_depends = ['dep1',
                            'dep2']
        # Mock _read_dependency_files
        with mock.patch(
            'ubuntu_system_tests.host.debian._read_dependency_files'
        ) as mock_read_depends:
            mock_read_depends.return_value = {
                'app2.tests': 'dep1',
                'app3.tests': 'dep2'
            }
            extra_depends = debian.get_extra_dependencies(
                self.config_stack.get('tests_to_run').split()
            )
        self.assertCountEqual(extra_depends, expected_depends)

    def test_read_dependency_file(self):
        with fixtures.TempDir() as deps_dir:
            deps_file1 = open(os.path.join(deps_dir.path, 'deps1.json'), 'w')
            deps_file2 = open(os.path.join(deps_dir.path, 'deps2.json'), 'w')
            deps_json1 = {
                "test_dependencies": [
                    {"id": "app1.tests", "depends": "dep1"}
                ]
            }
            deps_json2 = {
                "test_dependencies": [
                    {"id": "app2.tests", "depends": "dep2"}
                ]
            }
            deps_file1.write(json.dumps(deps_json1))
            deps_file2.write(json.dumps(deps_json2))
            deps_file1.close()
            deps_file2.close()

            expected = {'app1.tests': 'dep1', 'app2.tests': 'dep2'}
            actual = debian._read_dependency_files(deps_dir.path)
            self.assertEqual(expected, actual)

    def test_extra_dependencies_for_test(self):
        test = 'app1.tests.test_module.TestClass.test_y'
        dependencies_map = {'app1.tests': 'dep1', 'app2.tests': 'dep2'}
        extra_depends = debian._extra_dependencies_for_test(
            test, dependencies_map
        )
        self.assertEqual(extra_depends, ['dep1'])

    def test_create_temp_control(self):
        with NamedTemporaryFile() as control:
            control.write("""Tests: systemtests
Depends: dep1,
         dep2,
         dep3,

Tests: someothertests
Depends: other1,
         other2,
         other3,""".encode('utf-8'))
            control.flush()
            temp_control = debian.create_temp_control_file(
                ['dep4', 'dep5'], control.name
            )
            with open(temp_control) as temp:
                tmp_control_content = temp.read()
                expected_content = """Tests: systemtests
Depends: dep1,
         dep2,
         dep3,
         dep4,
         dep5,

Tests: someothertests
Depends: other1,
         other2,
         other3,"""
                self.assertEqual(tmp_control_content, expected_content)
